import __future__

import os
import threading

from chunker.utils import FilePtr


class ParseTimeoutException(Exception):
    """
    Exception indicating that timeout occurs when parsing.
    """
    pass


class Parser(object):
    """
    Parse chunks from file or streaming.

    - In file mode, :attr:`total_length` can be left None and will be read from
      file system.
    - In streaming mode, you **must** provide :attr:`total_length` to specify the
      end of the stream.

    :param fp: File object to be read from.
    :param total_length: Total length of the source.
    :throws ParseTimeoutException: Timeout. Default is 60s.
    """
    ChunkClasses = ()
    Timeout = 60

    def __init__(self, fp, total_length=None):
        if isinstance(fp, FilePtr):
            self.fp = fp
        else:
            self.fp = FilePtr(fp, total_length)
        self.chunks = []
        self._timeout_timer = None
        self.is_timeout = False
        self.is_debug = False

    @staticmethod
    def get_file_length(fp):
        return os.fstat(fp.fileno()).st_size

    def parse(self):
        """
        Start parsing.

        If you want more debug info, call :meth:`enable_debug` before parsing.
        """
        self._set_timeout()

        while self.fp.tell() < self.fp.total_length:
            for chunk_cls in self.__class__.ChunkClasses:
                if chunk_cls.matches(self.fp):
                    chunk = chunk_cls(self.fp, self)
                    chunk.populate()
                    self.chunks.append(chunk)

                    if self.is_debug:
                        print(chunk)
                        print('Now at', hex(self.fp.tell()))
                    break

            if self.is_timeout:
                raise ParseTimeoutException()

        self._timeout_timer.cancel()

    def _set_timeout(self):
        def handler():
            self.is_timeout = True

        self._timeout_timer = threading.Timer(self.__class__.Timeout, handler)
        self._timeout_timer.start()

    def enable_debug(self):
        """
        Print debugging info when parsing.
        """
        self.is_debug = True

    def close(self):
        """
        Close the file object.
        """
        self.fp.close()


class FileParser(Parser):
    """
    Parse from file.

    :param path: File path.
    """
    def __init__(self, path):
        fp = open(path, 'rb')
        super(FileParser, self).__init__(fp)



import os
import struct


class Field(object):
    """
    Base field class.

    It's abstract and cannot be used directly.

    :param name: Field name.
    """
    def __init__(self, name):
        self.name = name
        self.value = None
        self.length = 0

    def populate(self, fp, chunk):
        """
        Populate this field.

        You should override this or it will raise :exc:`NotImplementedError`.
        """
        raise NotImplementedError('Field is abstract.')


class BytesField(Field):
    """
    Base class for fixed length fields.

    It reads fixed count of bytes and decode them using :meth:`struct.unpack`.

    :param name: Field name.
    :param fmt: Format to be passed to :meth:`struct.unpack`.
    :param length: Field length in bytes.
    """
    def __init__(self, name, fmt, length):
        super(BytesField, self).__init__(name)
        self.name = name
        self.fmt = fmt
        self.length = length

    def populate(self, fp, chunk):
        buf = fp.read(self.length)
        self.value = struct.unpack(self.fmt, buf)[0]


class UnsignedLongField(BytesField):
    """
    Read 4 bytes and convert them into an unsigned long value.

    :param name: Field name.
    :param big_endian: If should be decoded in big-endian bytes order. Default is :const:`False`.
    """
    def __init__(self, name, big_endian=False):
        endian = '>' if big_endian else '<'
        fmt = endian + 'L'
        super(UnsignedLongField, self).__init__(name, fmt, 4)


class UnsignedShortField(BytesField):
    """
    Read 2 bytes and convert them into an unsigned short value.

    :param name: Field name.
    :param big_endian: If should be decoded in big-endian bytes order. Default is :const:`False`.
    """
    def __init__(self, name, big_endian=False):
        endian = '>' if big_endian else '<'
        fmt = endian + 'H'
        super(UnsignedShortField, self).__init__(name, fmt, 2)


class UnsignedCharField(BytesField):
    """
    Read 1 byte and convert them into an unsigned char value.

    :param name: Field name.
    :param big_endian: If should be decoded in big-endian bytes order. Default is :const:`False`.
    """
    def __init__(self, name, big_endian=False):
        endian = '>' if big_endian else '<'
        fmt = endian + 'B'
        super(UnsignedCharField, self).__init__(name, fmt, 1)


class VariableLengthField(Field):
    """
    Base class for field with variable length. Its length is based on another field.

    When populating, it simply read length and doesn't move on :obj:`fp` or read any data.

    :param name: Field name.
    :param length_field_name: Field name in the same chunk indicating the length of this field.
    """
    def __init__(self, name, length_field_name):
        super(VariableLengthField, self).__init__(name)
        self.length_field_name = length_field_name

    def populate(self, fp, chunk):
        self.length = chunk.__getattr__(self.length_field_name)


class StringField(VariableLengthField):
    """
    Read data with variable length. Its length is based on another field.

    :param name: Field name.
    :param length_field_name: Field name in the same chunk indicating the length of this field.
    """
    def populate(self, fp, chunk):
        super(StringField, self).populate(fp, chunk)
        self.value = fp.read(self.length)


class SkipBasedOnLengthField(VariableLengthField):
    """
    Skip data with variable length. Its length is based on another field.

    :param name: Field name.
    :param length_field_name: Field name in the same chunk indicating the length of this field.
    """
    def populate(self, fp, chunk):
        super(SkipBasedOnLengthField, self).populate(fp, chunk)
        fp.seek(self.length, os.SEEK_CUR)


class SkipBasedOnCalcField(Field):
    """
    Skip data with variable length. Its length is calculated from :func:`calc_func`.

    :param name: Field name.
    :param calc_func: Function to calculate the length to skip. When populating, this :obj:`chunk` will be passed to it as a parameter.

    Example::

        Fields = (
            UnsignedLongField('data_length'),
            SkipBasedOnCalcField('skip', lambda c: 3 * c.data_length),
        )
    """
    def __init__(self, name, calc_func):
        super(SkipBasedOnCalcField, self).__init__(name)
        self.calc_func = calc_func

    def populate(self, fp, chunk):
        offset = self.calc_func(chunk)
        fp.seek(offset, os.SEEK_CUR)


class SkipToTheEndField(Field):
    """
    Skip data to the end.

    The end depends on :attr:`fp.total_length`.

    :param name: Field name.
    """
    def populate(self, fp, chunk):
        if chunk.parser is not None:
            fp.seek(0, os.SEEK_END)
        else:
            raise AttributeError('No parser is provided')



import os


class FilePtr:
    """
    A wrapper for file object to maintain some more states during parsing.

    - In file mode, :attr:`total_length` can be left None and will be read from
      file system.
    - In streaming mode, you **must** provide :attr:`total_length` to specify the
      end of the stream.

    :param fp: Original file object.
    :param total_length: Length of this.
    :param offset: Offset to the original file.
    """
    def __init__(self, fp, total_length=None, offset=0):
        self.fp = fp
        self.offset = offset
        if total_length is not None:
            self.total_length = total_length
        else:
            self.total_length = FilePtr.get_file_length(fp)

    @staticmethod
    def get_file_length(fp):
        """
        Get length of the file from local file system.
        """
        return os.fstat(fp.fileno()).st_size

    def read(self, n):
        """
        Read bytes.

        :param n: Count of bytes to read.
        :returns: Bytes.
        """
        return self.fp.read(n)

    def seek(self, pos, mode=os.SEEK_SET):
        """
        Seek position in sub file.

        :param pos: Position.
        :param mode: Seek mode.
        """
        if mode == os.SEEK_CUR:
            self.fp.seek(pos, os.SEEK_CUR)
        elif mode == os.SEEK_SET:
            self.fp.seek(self.offset + pos, os.SEEK_SET)
        elif mode == os.SEEK_END:
            self.fp.seek(self.offset + self.total_length + pos, os.SEEK_SET)

    def tell(self):
        """
        Current position in sub file.

        :returns: Current position.
        """
        return self.fp.tell() - self.offset

    def close(self):
        """
        Set internal references to None.

        Note that it will not close the original file object.
        """
        self.fp = None
        self.offset = 0
        self.total_length = 0

    def save_state(self):
        """
        Save internal state.
        """
        self._saved_pos = self.tell()

    def restore_state(self):
        """
        Restore last internal state.
        """
        self.seek(self._saved_pos, os.SEEK_SET)



import os
import struct
import sys
import unittest

sys.path.insert(0, os.path.dirname(os.path.realpath(__file__)) + '/..')

if sys.version_info.major == 3:
    from io import BytesIO
else:
    from StringIO import StringIO as BytesIO

from chunks import Chunk
from fields import UnsignedLongField, UnsignedShortField
from parsers import Parser, ParseTimeoutException
from utils import FilePtr


class OneFieldChunk(Chunk):
    Fields = (
        UnsignedLongField('long'),
    )

    @classmethod
    def matches(cls, fp):
        buf = fp.read(4)
        sig = struct.unpack('<L', buf)[0]
        if sig == 0x10101010:
            return True
        else:
            fp.seek(-4, os.SEEK_CUR)
            return False


class TwoFieldChunk(Chunk):
    Fields = (
        UnsignedLongField('long'),
        UnsignedShortField('short'),
    )

    @classmethod
    def matches(cls, fp):
        buf = fp.read(4)
        sig = struct.unpack('<L', buf)[0]
        if sig == 0x01020304:
            return True
        else:
            fp.seek(-4, os.SEEK_CUR)
            return False


class ChunkTest(unittest.TestCase):
    def setUp(self):
        self.fp = BytesIO()
        sig = 0x01020304
        self.fp.write(struct.pack('<L', sig))
        long = 10
        self.fp.write(struct.pack('<L', long))
        short = 5
        self.fp.write(struct.pack('<H', short))
        self.fp.seek(0)

        self.fake_fp = BytesIO()
        sig = 0x04030201
        self.fake_fp.write(struct.pack('<L', sig))
        self.fake_fp.seek(0)

    def test_match(self):
        self.assertTrue(TwoFieldChunk.matches(self.fp))
        self.assertEqual(self.fp.tell(), 4)

        self.assertFalse(TwoFieldChunk.matches(self.fake_fp))
        self.assertEqual(self.fake_fp.tell(), 0)

    def test_populate(self):
        self.assertTrue(TwoFieldChunk.matches(self.fp))
        c = TwoFieldChunk(self.fp, None)
        c.populate()
        self.assertEqual(c.long, 10)
        self.assertEqual(c.short, 5)

    def tearDown(self):
        self.fp.close()
        self.fake_fp.close()


class OneFieldChunkWithProtectedMatches(Chunk):
    Fields = (
        UnsignedLongField('signature'),
        UnsignedLongField('long'),
    )

    @classmethod
    def safe_matches(cls, fp):
        buf = fp.read(4)
        sig = struct.unpack('<L', buf)[0]
        if sig == 0x10101010:
            return True
        else:
            return False


class ChunkWithProtectedMatchesTest(unittest.TestCase):
    def setUp(self):
        fp = BytesIO()
        sig = 0x10101010
        fp.write(struct.pack('<L', sig))
        long = 10
        fp.write(struct.pack('<L', long))
        fp.seek(0)

        self.fp = FilePtr(fp, total_length=8)

    def test_populate(self):
        self.assertTrue(OneFieldChunkWithProtectedMatches.matches(self.fp))
        c = OneFieldChunkWithProtectedMatches(self.fp, None)
        c.populate()
        self.assertEqual(c.long, 10)


class TwoChunksParser(Parser):
    ChunkClasses = (
        TwoFieldChunk,
        OneFieldChunk,
    )


class QuickTimeoutParser(Parser):
    Timeout = 0


class ParserTest(unittest.TestCase):
    TempFile = 'test.tmp'

    def setUp(self):
        self.fp = open(ParserTest.TempFile, 'wb+')

        # Chunk 1
        sig = 0x01020304
        self.fp.write(struct.pack('<L', sig))
        long = 10
        self.fp.write(struct.pack('<L', long))
        short = 5
        self.fp.write(struct.pack('<H', short))

        # Chunk 2
        sig = 0x10101010
        self.fp.write(struct.pack('<L', sig))
        long = 15
        self.fp.write(struct.pack('<L', long))

        # Reset cursor
        self.fp.flush()
        self.fp.seek(0)

    def test_parse(self):
        parser = TwoChunksParser(self.fp)
        parser.parse()
        self.assertEqual(parser.chunks[0].long, 10)
        self.assertEqual(parser.chunks[0].short, 5)
        self.assertEqual(parser.chunks[1].long, 15)
        self.assertFalse(parser.is_timeout)

    def test_timeout(self):
        with self.assertRaises(ParseTimeoutException):
            parser = QuickTimeoutParser(self.fp)
            parser.parse()
            self.assertTrue(parser.is_timeout)

    def tearDown(self):
        self.fp.close()

        if os.path.exists(ParserTest.TempFile):
            os.remove(ParserTest.TempFile)


if __name__ == '__main__':
    unittest.main()



import __future__

import copy
import os


class Chunk(object):
    """
    Chunk in the file. A chunk usually contains a signature indicating its
    type, and a group of data fields.

    :param fp: File object to read from.
    :param parser: The parser it belongs to. It's optional and you can leave it None.

    To define your chunk class, you should follow these steps:

    1. Define its fields. Fields are populated in this order. See :doc:`fields` for more information.

    2. Define a :meth:`matches` or :meth:`safe_matches` class method to judge if the following bytes match this type of chunk.

        * Override :meth:`matches` to handle :attr:`fp` state by yourself.
        * Override :meth:`safe_matches` to get :attr:`fp` state auto saved.

    Example::

        import os
        import struct

        from chunker.chunks import Chunk

        class DataChunk(Chunk):
            Fields = (
                UnsignedLongField('sig'),   # 0x01020304
                UnsignedLongField('type'),
                UnsignedLongField('data_length'),
                StringField('data', length_field_name='data_length'),
            )

            @classmethod
            def safe_matches(fp):
                buf = fp.read(4)
                type = struct.unpack('>H', buf)[0]

                return type == 0x01020304

        if DataChunk.matches(fp):
            c = DataChunk(fp, None)
            print(c.data) # Access field value by its name
    """
    Fields = ()

    def __new__(cls, *args, **kargs):
        instance = super(Chunk, cls).__new__(cls)
        instance.fields = copy.deepcopy(cls.Fields)
        instance.fields_map = {f.name: f for f in instance.fields}
        return instance

    def __init__(self, fp, parser):
        self.fp = fp
        self.parser = parser

    def __getattr__(self, key):
        return self.fields_map[key].value

    @classmethod
    def matches(cls, fp):
        """
        Read next a few bytes to judge if the following data match this type
        of chunk.

        It calls :meth:`safe_matches` and restores :attr:`fp` state by default.

        You can override this to avoid saving :attr:`fp` state.

        :param fp: File object to read from.
        :returns: If the following bytes match this chunk type.
        """
        fp.save_state()
        matches = cls.safe_matches(fp)
        fp.restore_state()
        return matches

    @classmethod
    def safe_matches(cls, fp):
        """
        The difference between :meth:`safe_matches` and :meth:`matches` is that if you override this, you can get your :attr:`fp` state auto saved.

        You only need to override one of them:

        * Override :meth:`matches` to handle :attr:`fp` state by yourself.
        * Override :meth:`safe_matches` to get :attr:`fp` state auto saved.

        :param fp: File object to read from.
        :returns: If the following bytes match this chunk type.
        """
        return False

    def populate(self):
        """
        Populate chunk fields. After that, you can access each field data
        directly by field name.
        """
        for field in self.fields:
            field.populate(self.fp, self)
            # if self.parser is not None and self.parser.is_debug:
            #     print('%s: %s @ %s' % (
            #         field.name, field.value, hex(self.fp.tell())))

    def __str__(self):
        dumps = []
        dumps.append('----------')
        dumps.append(self.__class__.__name__)
        dumps.append('----------')
        for field in self.fields:
            dumps.append('%s: %s' % (field.name, field.value))
        dumps.append('==========')
        return '\n'.join(dumps)



#!/usr/bin/env python
from distutils.core import setup

setup(
    name='chunker',
    version='1.1',
    description='Easy Chunk-Based File Structure Parser',
    author='Archangel_SDY',
    author_email='Archangel.SDY@gmail.com',
    url='http://archangelsdy.github.io/chunker/',
    packages=['chunker']
)
